body pre {
  margin: 0.5em 10% 0.5em 1em;
  line-height: 1.0;
  color: navy;
}
dl {
  margin: .2em 0;
  line-height: 1.2;
}

dt {
  margin-top: 0.1em;
}

span.man-highlight { background: yellow; }

a.st-desc {
	font-weight: bold;
	color: black;
	text-decoration: none;
}

a.st-desc:visited {
	font-weight: bold;
	color: black;
	text-decoration: none;
}

a.st-desc:hover {
	text-decoration: underline;
}

a.st-synopsis {
	font-weight: bold;
	color: black;
	text-decoration: none;
}

a.st-synopsis:visited {
	font-weight: bold;
	color: black;
	text-decoration: none;
}

a.st-synopsis:hover {
	text-decoration: underline;
}
