/* libcoap unit tests
 *
 * Copyright (C) 2013 Olaf Bergmann <bergmann@tzi.org>
 *
 * SPDX-License-Identifier: BSD-2-Clause
 *
 * This file is part of the CoAP library libcoap. Please see
 * README for terms of use.
 */

#include <CUnit/CUnit.h>

CU_pSuite t_init_error_response_tests(void);
